@@ -10,40 +10,48 @@
 # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
 # ANY KIND, either express or implied. See the License for the specific
 # language governing permissions and limitations under the License.
+import io
 import os.path
+import random
+import re
 import shutil
 import tempfile
 import threading
-import random
 import time
+from io import BytesIO, StringIO
 
-from tests import unittest
-from tests import RecordingSubscriber
-from s3transfer.compat import six
-from s3transfer.futures import TransferFuture
-from s3transfer.futures import TransferMeta
-from s3transfer.utils import get_callbacks
-from s3transfer.utils import random_file_extension
-from s3transfer.utils import invoke_progress_callbacks
-from s3transfer.utils import calculate_range_parameter
-from s3transfer.utils import CallArgs
-from s3transfer.utils import FunctionContainer
-from s3transfer.utils import CountCallbackInvoker
-from s3transfer.utils import OSUtils
-from s3transfer.utils import DeferredOpenFile
-from s3transfer.utils import ReadFileChunk
-from s3transfer.utils import StreamReaderProgress
-from s3transfer.utils import TaskSemaphore
-from s3transfer.utils import SlidingWindowSemaphore
-from s3transfer.utils import NoResourcesAvailable
+from s3transfer.futures import TransferFuture, TransferMeta
+from s3transfer.utils import (
+    MAX_PARTS,
+    MAX_SINGLE_UPLOAD_SIZE,
+    MIN_UPLOAD_CHUNKSIZE,
+    CallArgs,
+    ChunksizeAdjuster,
+    CountCallbackInvoker,
+    DeferredOpenFile,
+    FunctionContainer,
+    NoResourcesAvailable,
+    OSUtils,
+    ReadFileChunk,
+    SlidingWindowSemaphore,
+    StreamReaderProgress,
+    TaskSemaphore,
+    calculate_num_parts,
+    calculate_range_parameter,
+    get_callbacks,
+    get_filtered_dict,
+    invoke_progress_callbacks,
+    random_file_extension,
+)
+from tests import NonSeekableWriter, RecordingSubscriber, mock, unittest
 
 
 class TestGetCallbacks(unittest.TestCase):
     def setUp(self):
         self.subscriber = RecordingSubscriber()
         self.second_subscriber = RecordingSubscriber()
-        self.call_args = CallArgs(subscribers=[
-            self.subscriber, self.second_subscriber]
+        self.call_args = CallArgs(
+            subscribers=[self.subscriber, self.second_subscriber]
         )
         self.transfer_meta = TransferMeta(self.call_args)
         self.transfer_future = TransferFuture(self.transfer_meta)
@@ -58,8 +66,7 @@ class TestGetCallbacks(unittest.TestCase):
         # one of them and checking that the future was used in the call.
         callbacks[0]()
         self.assertEqual(
-            self.subscriber.on_queued_calls,
-            [{'future': self.transfer_future}]
+            self.subscriber.on_queued_calls, [{'future': self.transfer_future}]
         )
 
     def test_get_callbacks_for_missing_type(self):
@@ -69,6 +76,15 @@ class TestGetCallbacks(unittest.TestCase):
         self.assertEqual(len(callbacks), 0)
 
 
+class TestGetFilteredDict(unittest.TestCase):
+    def test_get_filtered_dict(self):
+        original = {'Include': 'IncludeValue', 'NotInlude': 'NotIncludeValue'}
+        whitelist = ['Include']
+        self.assertEqual(
+            get_filtered_dict(original, whitelist), {'Include': 'IncludeValue'}
+        )
+
+
 class TestCallArgs(unittest.TestCase):
     def test_call_args(self):
         call_args = CallArgs(foo='bar', biz='baz')
@@ -82,15 +98,20 @@ class TestFunctionContainer(unittest.TestCase):
 
     def test_call(self):
         func_container = FunctionContainer(
-            self.get_args_kwargs, 'foo', bar='baz')
+            self.get_args_kwargs, 'foo', bar='baz'
+        )
         self.assertEqual(func_container(), (('foo',), {'bar': 'baz'}))
 
     def test_repr(self):
         func_container = FunctionContainer(
-            self.get_args_kwargs, 'foo', bar='baz')
+            self.get_args_kwargs, 'foo', bar='baz'
+        )
         self.assertEqual(
-            str(func_container), 'Function: %s with args %s and kwargs %s' % (
-                self.get_args_kwargs, ('foo',), {'bar': 'baz'}))
+            str(func_container),
+            'Function: {} with args {} and kwargs {}'.format(
+                self.get_args_kwargs, ('foo',), {'bar': 'baz'}
+            ),
+        )
 
 
 class TestCountCallbackInvoker(unittest.TestCase):
@@ -146,8 +167,7 @@ class TestCountCallbackInvoker(unittest.TestCase):
 
 class TestRandomFileExtension(unittest.TestCase):
     def test_has_proper_length(self):
-        self.assertEqual(
-            len(random_file_extension(num_digits=4)), 4)
+        self.assertEqual(len(random_file_extension(num_digits=4)), 4)
 
 
 class TestInvokeProgressCallbacks(unittest.TestCase):
@@ -162,6 +182,14 @@ class TestInvokeProgressCallbacks(unittest.TestCase):
         self.assertEqual(len(recording_subscriber.on_progress_calls), 0)
 
 
+class TestCalculateNumParts(unittest.TestCase):
+    def test_calculate_num_parts_divisible(self):
+        self.assertEqual(calculate_num_parts(size=4, part_size=2), 2)
+
+    def test_calculate_num_parts_not_divisible(self):
+        self.assertEqual(calculate_num_parts(size=3, part_size=2), 2)
+
+
 class TestCalculateRangeParameter(unittest.TestCase):
     def setUp(self):
         self.part_size = 5
@@ -170,17 +198,20 @@ class TestCalculateRangeParameter(unittest.TestCase):
 
     def test_calculate_range_paramter(self):
         range_val = calculate_range_parameter(
-            self.part_size, self.part_index, self.num_parts)
+            self.part_size, self.part_index, self.num_parts
+        )
         self.assertEqual(range_val, 'bytes=5-9')
 
     def test_last_part_with_no_total_size(self):
         range_val = calculate_range_parameter(
-            self.part_size, self.part_index, num_parts=2)
+            self.part_size, self.part_index, num_parts=2
+        )
         self.assertEqual(range_val, 'bytes=5-')
 
     def test_last_part_with_total_size(self):
         range_val = calculate_range_parameter(
-            self.part_size, self.part_index, num_parts=2, total_size=8)
+            self.part_size, self.part_index, num_parts=2, total_size=8
+        )
         self.assertEqual(range_val, 'bytes=5-7')
 
 
@@ -192,6 +223,7 @@ class BaseUtilsTest(unittest.TestCase):
         with open(self.filename, 'wb') as f:
             f.write(self.content)
         self.amounts_seen = []
+        self.num_close_callback_calls = 0
 
     def tearDown(self):
         shutil.rmtree(self.tempdir)
@@ -199,15 +231,20 @@ class BaseUtilsTest(unittest.TestCase):
     def callback(self, bytes_transferred):
         self.amounts_seen.append(bytes_transferred)
 
+    def close_callback(self):
+        self.num_close_callback_calls += 1
+
 
 class TestOSUtils(BaseUtilsTest):
     def test_get_file_size(self):
         self.assertEqual(
-            OSUtils().get_file_size(self.filename), len(self.content))
+            OSUtils().get_file_size(self.filename), len(self.content)
+        )
 
     def test_open_file_chunk_reader(self):
         reader = OSUtils().open_file_chunk_reader(
-            self.filename, 0, 3, [self.callback])
+            self.filename, 0, 3, [self.callback]
+        )
 
         # The returned reader should be a ReadFileChunk.
         self.assertIsInstance(reader, ReadFileChunk)
@@ -219,14 +256,17 @@ class TestOSUtils(BaseUtilsTest):
     def test_open_file_chunk_reader_from_fileobj(self):
         with open(self.filename, 'rb') as f:
             reader = OSUtils().open_file_chunk_reader_from_fileobj(
-                f, len(self.content), len(self.content), [self.callback])
+                f, len(self.content), len(self.content), [self.callback]
+            )
 
             # The returned reader should be a ReadFileChunk.
             self.assertIsInstance(reader, ReadFileChunk)
             # The content of the reader should be correct.
             self.assertEqual(reader.read(), self.content)
+            reader.close()
             # Callbacks should be disabled depspite being passed in.
             self.assertEqual(self.amounts_seen, [])
+            self.assertEqual(self.num_close_callback_calls, 0)
 
     def test_open_file(self):
         fileobj = OSUtils().open(os.path.join(self.tempdir, 'foo'), 'w')
@@ -251,58 +291,168 @@ class TestOSUtils(BaseUtilsTest):
         self.assertFalse(os.path.exists(self.filename))
         self.assertTrue(os.path.exists(new_filename))
 
+    def test_is_special_file_for_normal_file(self):
+        self.assertFalse(OSUtils().is_special_file(self.filename))
 
-class TestDefferedOpenFile(BaseUtilsTest):
+    def test_is_special_file_for_non_existant_file(self):
+        non_existant_filename = os.path.join(self.tempdir, 'no-exist')
+        self.assertFalse(os.path.exists(non_existant_filename))
+        self.assertFalse(OSUtils().is_special_file(non_existant_filename))
+
+    def test_get_temp_filename(self):
+        filename = 'myfile'
+        self.assertIsNotNone(
+            re.match(
+                r'%s\.[0-9A-Fa-f]{8}$' % filename,
+                OSUtils().get_temp_filename(filename),
+            )
+        )
+
+    def test_get_temp_filename_len_255(self):
+        filename = 'a' * 255
+        temp_filename = OSUtils().get_temp_filename(filename)
+        self.assertLessEqual(len(temp_filename), 255)
+
+    def test_get_temp_filename_len_gt_255(self):
+        filename = 'a' * 280
+        temp_filename = OSUtils().get_temp_filename(filename)
+        self.assertLessEqual(len(temp_filename), 255)
+
+    def test_allocate(self):
+        truncate_size = 1
+        OSUtils().allocate(self.filename, truncate_size)
+        with open(self.filename, 'rb') as f:
+            self.assertEqual(len(f.read()), truncate_size)
+
+    @mock.patch('s3transfer.utils.fallocate')
+    def test_allocate_with_io_error(self, mock_fallocate):
+        mock_fallocate.side_effect = IOError()
+        with self.assertRaises(IOError):
+            OSUtils().allocate(self.filename, 1)
+        self.assertFalse(os.path.exists(self.filename))
+
+    @mock.patch('s3transfer.utils.fallocate')
+    def test_allocate_with_os_error(self, mock_fallocate):
+        mock_fallocate.side_effect = OSError()
+        with self.assertRaises(OSError):
+            OSUtils().allocate(self.filename, 1)
+        self.assertFalse(os.path.exists(self.filename))
+
+
+class TestDeferredOpenFile(BaseUtilsTest):
     def setUp(self):
-        super(TestDefferedOpenFile, self).setUp()
+        super().setUp()
         self.filename = os.path.join(self.tempdir, 'foo')
         self.contents = b'my contents'
         with open(self.filename, 'wb') as f:
             f.write(self.contents)
-        self.deferred_open_file = DeferredOpenFile(self.filename)
-        self.open_called_count = 0
-        self.deferred_open_file.OPEN_METHOD = self.counting_open_method
+        self.deferred_open_file = DeferredOpenFile(
+            self.filename, open_function=self.recording_open_function
+        )
+        self.open_call_args = []
 
     def tearDown(self):
         self.deferred_open_file.close()
-        super(TestDefferedOpenFile, self).tearDown()
+        super().tearDown()
 
-    def counting_open_method(self, filename, mode):
-        self.open_called_count += 1
+    def recording_open_function(self, filename, mode):
+        self.open_call_args.append((filename, mode))
         return open(filename, mode)
 
+    def open_nonseekable(self, filename, mode):
+        self.open_call_args.append((filename, mode))
+        return NonSeekableWriter(BytesIO(self.content))
+
     def test_instantiation_does_not_open_file(self):
-        deferred_open_file = DeferredOpenFile(self.filename)
-        self.open_called_count = 0
-        deferred_open_file.OPEN_METHOD = self.counting_open_method
-        self.assertEqual(self.open_called_count, 0)
+        DeferredOpenFile(
+            self.filename, open_function=self.recording_open_function
+        )
+        self.assertEqual(len(self.open_call_args), 0)
+
+    def test_name(self):
+        self.assertEqual(self.deferred_open_file.name, self.filename)
 
     def test_read(self):
         content = self.deferred_open_file.read(2)
         self.assertEqual(content, self.contents[0:2])
         content = self.deferred_open_file.read(2)
         self.assertEqual(content, self.contents[2:4])
-        self.assertEqual(self.open_called_count, 1)
+        self.assertEqual(len(self.open_call_args), 1)
+
+    def test_write(self):
+        self.deferred_open_file = DeferredOpenFile(
+            self.filename,
+            mode='wb',
+            open_function=self.recording_open_function,
+        )
+
+        write_content = b'foo'
+        self.deferred_open_file.write(write_content)
+        self.deferred_open_file.write(write_content)
+        self.deferred_open_file.close()
+        # Both of the writes should now be in the file.
+        with open(self.filename, 'rb') as f:
+            self.assertEqual(f.read(), write_content * 2)
+        # Open should have only been called once.
+        self.assertEqual(len(self.open_call_args), 1)
 
     def test_seek(self):
         self.deferred_open_file.seek(2)
         content = self.deferred_open_file.read(2)
         self.assertEqual(content, self.contents[2:4])
-        self.assertEqual(self.open_called_count, 1)
+        self.assertEqual(len(self.open_call_args), 1)
+
+    def test_open_does_not_seek_with_zero_start_byte(self):
+        self.deferred_open_file = DeferredOpenFile(
+            self.filename,
+            mode='wb',
+            start_byte=0,
+            open_function=self.open_nonseekable,
+        )
+
+        try:
+            # If this seeks, an UnsupportedOperation error will be raised.
+            self.deferred_open_file.write(b'data')
+        except io.UnsupportedOperation:
+            self.fail('DeferredOpenFile seeked upon opening')
+
+    def test_open_seeks_with_nonzero_start_byte(self):
+        self.deferred_open_file = DeferredOpenFile(
+            self.filename,
+            mode='wb',
+            start_byte=5,
+            open_function=self.open_nonseekable,
+        )
+
+        # Since a non-seekable file is being opened, calling Seek will raise
+        # an UnsupportedOperation error.
+        with self.assertRaises(io.UnsupportedOperation):
+            self.deferred_open_file.write(b'data')
 
     def test_tell(self):
         self.deferred_open_file.tell()
         # tell() should not have opened the file if it has not been seeked
         # or read because we know the start bytes upfront.
-        self.assertEqual(self.open_called_count, 0)
+        self.assertEqual(len(self.open_call_args), 0)
 
         self.deferred_open_file.seek(2)
         self.assertEqual(self.deferred_open_file.tell(), 2)
-        self.assertEqual(self.open_called_count, 1)
+        self.assertEqual(len(self.open_call_args), 1)
+
+    def test_open_args(self):
+        self.deferred_open_file = DeferredOpenFile(
+            self.filename,
+            mode='ab+',
+            open_function=self.recording_open_function,
+        )
+        # Force an open
+        self.deferred_open_file.write(b'data')
+        self.assertEqual(len(self.open_call_args), 1)
+        self.assertEqual(self.open_call_args[0], (self.filename, 'ab+'))
 
     def test_context_handler(self):
         with self.deferred_open_file:
-            self.assertEqual(self.open_called_count, 1)
+            self.assertEqual(len(self.open_call_args), 1)
 
 
 class TestReadFileChunk(BaseUtilsTest):
@@ -311,7 +461,8 @@ class TestReadFileChunk(BaseUtilsTest):
         with open(filename, 'wb') as f:
             f.write(b'onetwothreefourfivesixseveneightnineten')
         chunk = ReadFileChunk.from_filename(
-            filename, start_byte=0, chunk_size=3)
+            filename, start_byte=0, chunk_size=3
+        )
         self.assertEqual(chunk.read(), b'one')
         self.assertEqual(chunk.read(), b'')
 
@@ -320,7 +471,8 @@ class TestReadFileChunk(BaseUtilsTest):
         with open(filename, 'wb') as f:
             f.write(b'onetwothreefourfivesixseveneightnineten')
         chunk = ReadFileChunk.from_filename(
-            filename, start_byte=11, chunk_size=4)
+            filename, start_byte=11, chunk_size=4
+        )
         self.assertEqual(chunk.read(1), b'f')
         self.assertEqual(chunk.read(1), b'o')
         self.assertEqual(chunk.read(1), b'u')
@@ -332,7 +484,8 @@ class TestReadFileChunk(BaseUtilsTest):
         with open(filename, 'wb') as f:
             f.write(b'onetwothreefourfivesixseveneightnineten')
         chunk = ReadFileChunk.from_filename(
-            filename, start_byte=11, chunk_size=4)
+            filename, start_byte=11, chunk_size=4
+        )
         self.assertEqual(chunk.read(), b'four')
         chunk.seek(0)
         self.assertEqual(chunk.read(), b'four')
@@ -342,7 +495,8 @@ class TestReadFileChunk(BaseUtilsTest):
         with open(filename, 'wb') as f:
             f.write(b'onetwothreefourfivesixseveneightnineten')
         chunk = ReadFileChunk.from_filename(
-            filename, start_byte=36, chunk_size=100000)
+            filename, start_byte=36, chunk_size=100000
+        )
         self.assertEqual(chunk.read(), b'ten')
         self.assertEqual(chunk.read(), b'')
         self.assertEqual(len(chunk), 3)
@@ -352,20 +506,136 @@ class TestReadFileChunk(BaseUtilsTest):
         with open(filename, 'wb') as f:
             f.write(b'onetwothreefourfivesixseveneightnineten')
         chunk = ReadFileChunk.from_filename(
-            filename, start_byte=36, chunk_size=100000)
+            filename, start_byte=36, chunk_size=100000
+        )
         self.assertEqual(chunk.tell(), 0)
         self.assertEqual(chunk.read(), b'ten')
         self.assertEqual(chunk.tell(), 3)
         chunk.seek(0)
         self.assertEqual(chunk.tell(), 0)
+        chunk.seek(1, whence=1)
+        self.assertEqual(chunk.tell(), 1)
+        chunk.seek(-1, whence=1)
+        self.assertEqual(chunk.tell(), 0)
+        chunk.seek(-1, whence=2)
+        self.assertEqual(chunk.tell(), 2)
+
+    def test_tell_and_seek_boundaries(self):
+        # Test to ensure ReadFileChunk behaves the same as the
+        # Python standard library around seeking and reading out
+        # of bounds in a file object.
+        data = b'abcdefghij12345678klmnopqrst'
+        start_pos = 10
+        chunk_size = 8
+
+        # Create test file
+        filename = os.path.join(self.tempdir, 'foo')
+        with open(filename, 'wb') as f:
+            f.write(data)
+
+        # ReadFileChunk should be a substring of only numbers
+        file_objects = [
+            ReadFileChunk.from_filename(
+                filename, start_byte=start_pos, chunk_size=chunk_size
+            )
+        ]
+
+        # Uncomment next line to validate we match Python's io.BytesIO
+        # file_objects.append(io.BytesIO(data[start_pos:start_pos+chunk_size]))
+
+        for obj in file_objects:
+            self._assert_whence_start_behavior(obj)
+            self._assert_whence_end_behavior(obj)
+            self._assert_whence_relative_behavior(obj)
+            self._assert_boundary_behavior(obj)
+
+    def _assert_whence_start_behavior(self, file_obj):
+        self.assertEqual(file_obj.tell(), 0)
+
+        file_obj.seek(1, 0)
+        self.assertEqual(file_obj.tell(), 1)
+
+        file_obj.seek(1)
+        self.assertEqual(file_obj.tell(), 1)
+        self.assertEqual(file_obj.read(), b'2345678')
+
+        file_obj.seek(3, 0)
+        self.assertEqual(file_obj.tell(), 3)
+
+        file_obj.seek(0, 0)
+        self.assertEqual(file_obj.tell(), 0)
+
+    def _assert_whence_relative_behavior(self, file_obj):
+        self.assertEqual(file_obj.tell(), 0)
+
+        file_obj.seek(2, 1)
+        self.assertEqual(file_obj.tell(), 2)
+
+        file_obj.seek(1, 1)
+        self.assertEqual(file_obj.tell(), 3)
+        self.assertEqual(file_obj.read(), b'45678')
+
+        file_obj.seek(20, 1)
+        self.assertEqual(file_obj.tell(), 28)
+
+        file_obj.seek(-30, 1)
+        self.assertEqual(file_obj.tell(), 0)
+        self.assertEqual(file_obj.read(), b'12345678')
+
+        file_obj.seek(-8, 1)
+        self.assertEqual(file_obj.tell(), 0)
+
+    def _assert_whence_end_behavior(self, file_obj):
+        self.assertEqual(file_obj.tell(), 0)
+
+        file_obj.seek(-1, 2)
+        self.assertEqual(file_obj.tell(), 7)
+
+        file_obj.seek(1, 2)
+        self.assertEqual(file_obj.tell(), 9)
+
+        file_obj.seek(3, 2)
+        self.assertEqual(file_obj.tell(), 11)
+        self.assertEqual(file_obj.read(), b'')
+
+        file_obj.seek(-15, 2)
+        self.assertEqual(file_obj.tell(), 0)
+        self.assertEqual(file_obj.read(), b'12345678')
+
+        file_obj.seek(-8, 2)
+        self.assertEqual(file_obj.tell(), 0)
+
+    def _assert_boundary_behavior(self, file_obj):
+        # Verify we're at the start
+        self.assertEqual(file_obj.tell(), 0)
+
+        # Verify we can't move backwards beyond start of file
+        file_obj.seek(-10, 1)
+        self.assertEqual(file_obj.tell(), 0)
+
+        # Verify we *can* move after end of file, but return nothing
+        file_obj.seek(10, 2)
+        self.assertEqual(file_obj.tell(), 18)
+        self.assertEqual(file_obj.read(), b'')
+        self.assertEqual(file_obj.read(10), b'')
+
+        # Verify we can partially rewind
+        file_obj.seek(-12, 1)
+        self.assertEqual(file_obj.tell(), 6)
+        self.assertEqual(file_obj.read(), b'78')
+        self.assertEqual(file_obj.tell(), 8)
+
+        # Verify we can rewind to start
+        file_obj.seek(0)
+        self.assertEqual(file_obj.tell(), 0)
 
     def test_file_chunk_supports_context_manager(self):
         filename = os.path.join(self.tempdir, 'foo')
         with open(filename, 'wb') as f:
             f.write(b'abc')
-        with ReadFileChunk.from_filename(filename,
-                                         start_byte=0,
-                                         chunk_size=2) as chunk:
+        with ReadFileChunk.from_filename(
+            filename, start_byte=0, chunk_size=2
+        ) as chunk:
             val = chunk.read()
             self.assertEqual(val, b'ab')
 
@@ -375,13 +645,17 @@ class TestReadFileChunk(BaseUtilsTest):
         filename = os.path.join(self.tempdir, 'foo')
         open(filename, 'wb').close()
         chunk = ReadFileChunk.from_filename(
-            filename, start_byte=0, chunk_size=10)
+            filename, start_byte=0, chunk_size=10
+        )
         self.assertEqual(list(chunk), [])
 
     def test_callback_is_invoked_on_read(self):
         chunk = ReadFileChunk.from_filename(
-            self.filename, start_byte=0, chunk_size=3,
-            callbacks=[self.callback])
+            self.filename,
+            start_byte=0,
+            chunk_size=3,
+            callbacks=[self.callback],
+        )
         chunk.read(1)
         chunk.read(1)
         chunk.read(1)
@@ -389,8 +663,11 @@ class TestReadFileChunk(BaseUtilsTest):
 
     def test_all_callbacks_invoked_on_read(self):
         chunk = ReadFileChunk.from_filename(
-            self.filename, start_byte=0, chunk_size=3,
-            callbacks=[self.callback, self.callback])
+            self.filename,
+            start_byte=0,
+            chunk_size=3,
+            callbacks=[self.callback, self.callback],
+        )
         chunk.read(1)
         chunk.read(1)
         chunk.read(1)
@@ -400,8 +677,11 @@ class TestReadFileChunk(BaseUtilsTest):
 
     def test_callback_can_be_disabled(self):
         chunk = ReadFileChunk.from_filename(
-            self.filename, start_byte=0, chunk_size=3,
-            callbacks=[self.callback])
+            self.filename,
+            start_byte=0,
+            chunk_size=3,
+            callbacks=[self.callback],
+        )
         chunk.disable_callback()
         # Now reading from the ReadFileChunk should not invoke
         # the callback.
@@ -410,8 +690,11 @@ class TestReadFileChunk(BaseUtilsTest):
 
     def test_callback_will_also_be_triggered_by_seek(self):
         chunk = ReadFileChunk.from_filename(
-            self.filename, start_byte=0, chunk_size=3,
-            callbacks=[self.callback])
+            self.filename,
+            start_byte=0,
+            chunk_size=3,
+            callbacks=[self.callback],
+        )
         chunk.read(2)
         chunk.seek(0)
         chunk.read(2)
@@ -419,17 +702,189 @@ class TestReadFileChunk(BaseUtilsTest):
         chunk.read(2)
         self.assertEqual(self.amounts_seen, [2, -2, 2, -1, 2])
 
+    def test_callback_triggered_by_out_of_bound_seeks(self):
+        data = b'abcdefghij1234567890klmnopqr'
+
+        # Create test file
+        filename = os.path.join(self.tempdir, 'foo')
+        with open(filename, 'wb') as f:
+            f.write(data)
+        chunk = ReadFileChunk.from_filename(
+            filename, start_byte=10, chunk_size=10, callbacks=[self.callback]
+        )
+
+        # Seek calls that generate "0" progress are skipped by
+        # invoke_progress_callbacks and won't appear in the list.
+        expected_callback_prog = [10, -5, 5, -1, 1, -1, 1, -5, 5, -10]
+
+        self._assert_out_of_bound_start_seek(chunk, expected_callback_prog)
+        self._assert_out_of_bound_relative_seek(chunk, expected_callback_prog)
+        self._assert_out_of_bound_end_seek(chunk, expected_callback_prog)
+
+    def _assert_out_of_bound_start_seek(self, chunk, expected):
+        # clear amounts_seen
+        self.amounts_seen = []
+        self.assertEqual(self.amounts_seen, [])
+
+        # (position, change)
+        chunk.seek(20)  # (20, 10)
+        chunk.seek(5)  # (5, -5)
+        chunk.seek(20)  # (20, 5)
+        chunk.seek(9)  # (9, -1)
+        chunk.seek(20)  # (20, 1)
+        chunk.seek(11)  # (11, 0)
+        chunk.seek(20)  # (20, 0)
+        chunk.seek(9)  # (9, -1)
+        chunk.seek(20)  # (20, 1)
+        chunk.seek(5)  # (5, -5)
+        chunk.seek(20)  # (20, 5)
+        chunk.seek(0)  # (0, -10)
+        chunk.seek(0)  # (0, 0)
+
+        self.assertEqual(self.amounts_seen, expected)
+
+    def _assert_out_of_bound_relative_seek(self, chunk, expected):
+        # clear amounts_seen
+        self.amounts_seen = []
+        self.assertEqual(self.amounts_seen, [])
+
+        # (position, change)
+        chunk.seek(20, 1)  # (20, 10)
+        chunk.seek(-15, 1)  # (5, -5)
+        chunk.seek(15, 1)  # (20, 5)
+        chunk.seek(-11, 1)  # (9, -1)
+        chunk.seek(11, 1)  # (20, 1)
+        chunk.seek(-9, 1)  # (11, 0)
+        chunk.seek(9, 1)  # (20, 0)
+        chunk.seek(-11, 1)  # (9, -1)
+        chunk.seek(11, 1)  # (20, 1)
+        chunk.seek(-15, 1)  # (5, -5)
+        chunk.seek(15, 1)  # (20, 5)
+        chunk.seek(-20, 1)  # (0, -10)
+        chunk.seek(-1000, 1)  # (0, 0)
+
+        self.assertEqual(self.amounts_seen, expected)
+
+    def _assert_out_of_bound_end_seek(self, chunk, expected):
+        # clear amounts_seen
+        self.amounts_seen = []
+        self.assertEqual(self.amounts_seen, [])
+
+        # (position, change)
+        chunk.seek(10, 2)  # (20, 10)
+        chunk.seek(-5, 2)  # (5, -5)
+        chunk.seek(10, 2)  # (20, 5)
+        chunk.seek(-1, 2)  # (9, -1)
+        chunk.seek(10, 2)  # (20, 1)
+        chunk.seek(1, 2)  # (11, 0)
+        chunk.seek(10, 2)  # (20, 0)
+        chunk.seek(-1, 2)  # (9, -1)
+        chunk.seek(10, 2)  # (20, 1)
+        chunk.seek(-5, 2)  # (5, -5)
+        chunk.seek(10, 2)  # (20, 5)
+        chunk.seek(-10, 2)  # (0, -10)
+        chunk.seek(-1000, 2)  # (0, 0)
+
+        self.assertEqual(self.amounts_seen, expected)
+
+    def test_close_callbacks(self):
+        with open(self.filename) as f:
+            chunk = ReadFileChunk(
+                f,
+                chunk_size=1,
+                full_file_size=3,
+                close_callbacks=[self.close_callback],
+            )
+            chunk.close()
+            self.assertEqual(self.num_close_callback_calls, 1)
+
+    def test_close_callbacks_when_not_enabled(self):
+        with open(self.filename) as f:
+            chunk = ReadFileChunk(
+                f,
+                chunk_size=1,
+                full_file_size=3,
+                enable_callbacks=False,
+                close_callbacks=[self.close_callback],
+            )
+            chunk.close()
+            self.assertEqual(self.num_close_callback_calls, 0)
+
+    def test_close_callbacks_when_context_handler_is_used(self):
+        with open(self.filename) as f:
+            with ReadFileChunk(
+                f,
+                chunk_size=1,
+                full_file_size=3,
+                close_callbacks=[self.close_callback],
+            ) as chunk:
+                chunk.read(1)
+            self.assertEqual(self.num_close_callback_calls, 1)
+
+    def test_signal_transferring(self):
+        chunk = ReadFileChunk.from_filename(
+            self.filename,
+            start_byte=0,
+            chunk_size=3,
+            callbacks=[self.callback],
+        )
+        chunk.signal_not_transferring()
+        chunk.read(1)
+        self.assertEqual(self.amounts_seen, [])
+        chunk.signal_transferring()
+        chunk.read(1)
+        self.assertEqual(self.amounts_seen, [1])
+
+    def test_signal_transferring_to_underlying_fileobj(self):
+        underlying_stream = mock.Mock()
+        underlying_stream.tell.return_value = 0
+        chunk = ReadFileChunk(underlying_stream, 3, 3)
+        chunk.signal_transferring()
+        self.assertTrue(underlying_stream.signal_transferring.called)
+
+    def test_no_call_signal_transferring_to_underlying_fileobj(self):
+        underlying_stream = mock.Mock(io.RawIOBase)
+        underlying_stream.tell.return_value = 0
+        chunk = ReadFileChunk(underlying_stream, 3, 3)
+        try:
+            chunk.signal_transferring()
+        except AttributeError:
+            self.fail(
+                'The stream should not have tried to call signal_transferring '
+                'to the underlying stream.'
+            )
+
+    def test_signal_not_transferring_to_underlying_fileobj(self):
+        underlying_stream = mock.Mock()
+        underlying_stream.tell.return_value = 0
+        chunk = ReadFileChunk(underlying_stream, 3, 3)
+        chunk.signal_not_transferring()
+        self.assertTrue(underlying_stream.signal_not_transferring.called)
+
+    def test_no_call_signal_not_transferring_to_underlying_fileobj(self):
+        underlying_stream = mock.Mock(io.RawIOBase)
+        underlying_stream.tell.return_value = 0
+        chunk = ReadFileChunk(underlying_stream, 3, 3)
+        try:
+            chunk.signal_not_transferring()
+        except AttributeError:
+            self.fail(
+                'The stream should not have tried to call '
+                'signal_not_transferring to the underlying stream.'
+            )
+
 
 class TestStreamReaderProgress(BaseUtilsTest):
     def test_proxies_to_wrapped_stream(self):
-        original_stream = six.StringIO('foobarbaz')
+        original_stream = StringIO('foobarbaz')
         wrapped = StreamReaderProgress(original_stream)
         self.assertEqual(wrapped.read(), 'foobarbaz')
 
     def test_callback_invoked(self):
-        original_stream = six.StringIO('foobarbaz')
+        original_stream = StringIO('foobarbaz')
         wrapped = StreamReaderProgress(
-            original_stream, [self.callback, self.callback])
+            original_stream, [self.callback, self.callback]
+        )
         self.assertEqual(wrapped.read(), 'foobarbaz')
         self.assertEqual(self.amounts_seen, [9, 9])
 
@@ -681,5 +1136,54 @@ class TestThreadingPropertiesForSlidingWindowSemaphore(unittest.TestCase):
         # Should have all the available resources freed.
         self.assertEqual(sem.current_count(), 5)
         # Should have acquired num_threads * num_iterations
-        self.assertEqual(sem.acquire('a', blocking=False),
-                         num_threads * num_iterations)
+        self.assertEqual(
+            sem.acquire('a', blocking=False), num_threads * num_iterations
+        )
+
+
+class TestAdjustChunksize(unittest.TestCase):
+    def setUp(self):
+        self.adjuster = ChunksizeAdjuster()
+
+    def test_valid_chunksize(self):
+        chunksize = 7 * (1024**2)
+        file_size = 8 * (1024**2)
+        new_size = self.adjuster.adjust_chunksize(chunksize, file_size)
+        self.assertEqual(new_size, chunksize)
+
+    def test_chunksize_below_minimum(self):
+        chunksize = MIN_UPLOAD_CHUNKSIZE - 1
+        file_size = 3 * MIN_UPLOAD_CHUNKSIZE
+        new_size = self.adjuster.adjust_chunksize(chunksize, file_size)
+        self.assertEqual(new_size, MIN_UPLOAD_CHUNKSIZE)
+
+    def test_chunksize_above_maximum(self):
+        chunksize = MAX_SINGLE_UPLOAD_SIZE + 1
+        file_size = MAX_SINGLE_UPLOAD_SIZE * 2
+        new_size = self.adjuster.adjust_chunksize(chunksize, file_size)
+        self.assertEqual(new_size, MAX_SINGLE_UPLOAD_SIZE)
+
+    def test_chunksize_too_small(self):
+        chunksize = 7 * (1024**2)
+        file_size = 5 * (1024**4)
+        # If we try to upload a 5TB file, we'll need to use 896MB part
+        # sizes.
+        new_size = self.adjuster.adjust_chunksize(chunksize, file_size)
+        self.assertEqual(new_size, 896 * (1024**2))
+        num_parts = file_size / new_size
+        self.assertLessEqual(num_parts, MAX_PARTS)
+
+    def test_unknown_file_size_with_valid_chunksize(self):
+        chunksize = 7 * (1024**2)
+        new_size = self.adjuster.adjust_chunksize(chunksize)
+        self.assertEqual(new_size, chunksize)
+
+    def test_unknown_file_size_below_minimum(self):
+        chunksize = MIN_UPLOAD_CHUNKSIZE - 1
+        new_size = self.adjuster.adjust_chunksize(chunksize)
+        self.assertEqual(new_size, MIN_UPLOAD_CHUNKSIZE)
+
+    def test_unknown_file_size_above_maximum(self):
+        chunksize = MAX_SINGLE_UPLOAD_SIZE + 1
+        new_size = self.adjuster.adjust_chunksize(chunksize)
+        self.assertEqual(new_size, MAX_SINGLE_UPLOAD_SIZE)
